from math import ceil
from typing import Callable, Dict, Type, Union
import random


import numpy as np
import pandas as pd
from desdeo_problem.surrogatemodels.SurrogateModels import BaseRegressor, ModelError
from scipy.special import expit
from sklearn.metrics import mean_squared_error, mean_squared_log_error, r2_score

from desdeo_emo.EAs.BaseEA import BaseEA
from desdeo_emo.EAs.PPGA import PPGA

from desdeo_emo.othertools.plotlyanimate import animate_init_, animate_next_
from desdeo_emo.population.SurrogatePopulation import SurrogatePopulation
from desdeo_emo.recombination.evodn2_xover_mutation import EvoDN2Recombination
from desdeo_emo.surrogatemodels.Problem import surrogateProblem


def negative_r2_score(y_true, y_pred):
    return -r2_score(y_true, y_pred)


class EvoDN2(BaseRegressor):
    def __init__(
        self,
        num_subnets: int = 4,
        num_subsets: int = 4,
        max_layers: int = 4,
        max_nodes: int = 4,
        p_omit: float = 0.2,
        w_low: float = -5.0,
        w_high: float = 5.0,
        subsets: list = None,
        activation_function: str = "sigmoid",
        loss_function: str = "mse",
        training_algorithm: BaseEA = PPGA,
        pop_size: int = 500,
        model_selection_criterion: str = "min_error",
        verbose: int = 0,
    ):
        loss_functions = {
            "mse": mean_squared_error,
            "msle": mean_squared_log_error,
            "neg_r2": negative_r2_score,
        }
        # Model Hyperparameters
        self.num_subnets: int = num_subnets
        self.num_subsets: int = num_subsets
        self.max_layers: int = max_layers
        self.max_nodes: int = max_nodes
        self.p_omit: float = p_omit
        self.w_low: float = w_low
        self.w_high: float = w_high
        self.subsets: list = subsets
        self.activation_function: str = activation_function
        self.loss_function_str: str = loss_function
        self.loss_function: Callable = loss_functions[loss_function]
        # EA Parameters
        self.training_algorithm: BaseEA = training_algorithm
        self.pop_size: int = pop_size
        self.model_selection_criterion: str = model_selection_criterion
        self.verbose: int = verbose
        self.X: np.ndarray = None
        self.y: np.ndarray = None
        self.model_trained: bool = False
        # Model parameters
        self._subnets = None
        self._last_layer = None
        # Extras
        self.performance: Dict = {"RMSE": None, "R^2": None, "Complexity": None}
        self.model_population = None

    def fit(self, X: np.ndarray, y: np.ndarray):
        if isinstance(X, (pd.DataFrame, pd.Series)):
            X = X.values
        if isinstance(y, (pd.DataFrame, pd.Series)):
            y = y.values.reshape(-1, 1)
        if X.shape[0] != y.shape[0]:
            msg = (
                f"Ensure that the number of samples in X and y are the same"
                f"Number of samples in X = {X.shape[0]}"
                f"Number of samples in y = {y.shape[0]}"
            )
            raise ModelError(msg)
        self.X = X
        self.y = y
        if self.subsets is None:
            self.subsets = []

            # Create random subsets of decision variables for each subnet

            for i in range(self.num_subnets):
                n = random.randint(1, self.X.shape[1])
                self.subsets.append(random.sample(range(self.X.shape[1]), n))

            # Ensure that each decision variable is used as an input in at least one subnet
            for n in list(range(self.X.shape[1])):
                if not any(n in k for k in self.subsets):
                    self.subsets[random.randint(0, self.num_subnets - 1)].append(n)

        # Create problem
        problem = surrogateProblem(performance_evaluator=self._model_performance)
        problem.n_of_objectives = 2
        # Create Population
        initial_pop = self._create_individuals()
        population = SurrogatePopulation(
            problem, self.pop_size, initial_pop, None, None, None
        )
        # Do evolution
        evolver = self.training_algorithm(problem, initial_population=population)
        recombinator = EvoDN2Recombination(evolver=evolver)
        evolver.population.recombination = recombinator
        figure = animate_init_(evolver.population.objectives, filename="EvoDN2.html")
        while evolver.continue_evolution():
            evolver.iterate()
            figure = animate_next_(
                evolver.population.objectives,
                figure,
                filename="EvoDN2.html",
                generation=evolver._iteration_counter,
            )
        self.model_population = evolver.population
        # Selection
        self.select()
        self.model_trained = True

    def _model_performance(
        self,
        individuals: np.ndarray = None,
        X: np.ndarray = None,
        y_true: np.ndarray = None,
    ):
        if individuals is None and self.model_trained is False:
            msg = "Model has not been trained yet"
            raise ModelError(msg)
        if individuals is None:
            individuals = self._subnets
        if X is None:
            X = self.X
            y = self.y
        if len(individuals) > 1:
            loss = []
            complexity = []
            for individual in individuals:
                penultimate_y_pred, ind_complexity = self._feed_forward(individual, X)
                y_pred, _ = self._calculate_linear(penultimate_y_pred)
                loss.append(self.loss_function(y, y_pred))
                complexity.append(ind_complexity)
        else:
            penultimate_y_pred, ind_complexity = self._feed_forward(individuals, X)
            y_pred, _ = self._calculate_linear(penultimate_y_pred)
            loss = self.loss_function(y, y_pred)
            complexity = ind_complexity
        return np.asarray((loss, complexity)).T

    def _feed_forward(self, subnets, X):
        network_complexity = []
        penultimate_output = np.empty((X.shape[0], 0))
        for i, subnet in enumerate(subnets):
            # Get the input variables for the first layer
            feed = X[:, self.subsets[i]]
            subnet_complexity = 1
            for layer in subnet:
                # Calculate the dot product + bias
                out = np.dot(feed, layer[1:, :]) + layer[0]
                subnet_complexity = np.dot(subnet_complexity, np.abs(layer[1:, :]))
                feed = self.activate(out)
            network_complexity.append(np.sum(subnet_complexity))
            penultimate_output = np.hstack((penultimate_output, feed))

        complexity = np.sum(network_complexity)
        return penultimate_output, complexity

    def _calculate_linear(self, previous_layer_output):
        """ Calculate the final layer using LLSQ or

        Parameters
        ----------
        non_linear_layer : np.ndarray
            Output of the activation function

        Returns
        -------
        linear_layer : np.ndarray
            The optimized weight matrix of the upper part of the network
        predicted_values : np.ndarray
            The prediction of the model
        """

        linear_layer = None
        previous_layer_output = np.hstack(
            (np.ones((previous_layer_output.shape[0], 1)), previous_layer_output)
        )
        linear_solution = np.linalg.lstsq(previous_layer_output, self.y, rcond=None)
        linear_layer = linear_solution[0]
        y_pred = np.dot(previous_layer_output, linear_layer)
        return y_pred, linear_layer

    def activate(self, x):
        if self.activation_function == "sigmoid":
            return expit(x)
        elif self.activation_function == "relu":
            return np.maximum(x, 0)
        elif self.activation_function == "tanh":
            return np.tanh(x)
        else:
            msg = (
                f"Given activation function not recognized: {self.activation_function}"
                f"\nActivation function should be one of ['relu', 'sigmoid', 'tanh']"
            )
            raise ModelError(msg)

    def predict(self, X):
        penultimate_y_pred, _ = self._feed_forward(self.subnets, X)
        y_pred = (
            np.dot(penultimate_y_pred, self._last_layer[1:, :]) + self._last_layer[0]
        )
        return y_pred

    def select(self):
        if self.model_selection_criterion == "min_error":
            # Return the model with the lowest error
            selected = np.argmin(self.model_population.objectives[:, 0])
        else:
            raise ModelError("Selection criterion not recognized. Use 'min_error'.")
        self.subnets = self.model_population.individuals[selected]
        penultimate_y_pred, complexity = self._feed_forward(self.subnets, self.X)
        y_pred, linear_layer = self._calculate_linear(penultimate_y_pred)
        self._last_layer = linear_layer
        self.performance["RMSE"] = np.sqrt(mean_squared_error(self.y, y_pred))
        self.performance["R^2"] = r2_score(self.y, y_pred)
        self.performance["Complexity"] = complexity

    def _create_individuals(self):
        individuals = []
        for i in range(self.pop_size):
            nets = []
            for j in range(self.num_subnets):

                layers = []
                num_layers = np.random.randint(1, self.max_layers)
                in_nodes = len(self.subsets[j])

                for k in range(num_layers):
                    out_nodes = random.randint(2, self.max_nodes)
                    net = np.random.uniform(
                        self.w_low,
                        self.w_high,
                        size=(in_nodes, out_nodes),
                    )
                    # Randomly set some weights to zero
                    zeros = np.random.choice(
                        np.arange(net.size),
                        ceil(net.size * self.p_omit),
                    )
                    net.ravel()[zeros] = 0

                    # Add bias
                    net = np.insert(net, 0, 1, axis=0)
                    in_nodes = out_nodes
                    layers.append(net)

                nets.append(layers)

            individuals.append(nets)

        return np.asarray(individuals)


import numpy as np


def tour_select(fitness, tournament_size):
    """Tournament selection. Choose number of individuals to participate
    and select the one with the best fitness.

    Parameters
    ----------
    fitness : array_like
        An array of each individual's fitness.
    tournament_size : int
        Number of participants in the tournament.

    Returns
    -------
    int
        The index of the best individual.
    """
    aspirants = np.random.choice(len(fitness)-1, tournament_size, replace=False)
    chosen = []
    for ind in aspirants:
        chosen.append([ind, fitness[ind]])
    chosen.sort(key=lambda x: x[1])

    return chosen[0][0]



from desdeo_emo.population.Population import BasePopulation, Population


class SurrogatePopulation(Population, BasePopulation):
    def __init__(
        self, problem, pop_size: int, initial_pop, xover, mutation, recombination
    ):
        BasePopulation.__init__(self, problem, pop_size)
        self.add(initial_pop)
        self.xover = xover
        self.mutation = mutation
        self.recombination = recombination



from random import choice, sample
import numpy as np

# from pygmo import fast_non_dominated_sorting as nds
from desdeo_tools.utilities import fast_non_dominated_sort as nds
from desdeo_emo.EAs.BaseEA import BaseEA, eaError


class PPGA(BaseEA):
    """Predatory-Prey genetic algorithm.

    A population of prey signify the various models or solutions to the problem at hand.
    Weaker prey, i.e. bad models or solutions, are killed by predators.
    The predators and prey are placed in a lattice, in which they are free to roam.

    In each generation, each predator gets a certain number of turns to move about and
    hunt in its neighbourhood, killing the weaker prey, according to a fitness criteria.
    After this, each prey gets a certain number of moves to pursue a random walk and to
    reproduce with other prey. Each reproduction step generates two new prey from two
    parents, by crossing over their attributes and adding random mutations. After each
    prey has completed its move, the whole process starts again.

    As the weaker individuals get eliminated in each generation, the population as a
    whole becomes more fit, i.e. the individuals get closer to the true pareto-optimal
    solutions.

    If you have any questions about the code, please contact:

    Bhupinder Saini: bhupinder.s.saini@jyu.fi
    Project researcher at University of Jyväskylä.

    Parameters
    ----------
    population : object
        The population object


    Notes
    -----
    The algorithm has been created earlier in MATLAB, and this Python implementation
    has been using that code as a basis. See references [4] for the study during which
    the original MATLAB version was created.
    Python code has been written by Niko Rissanen under the supervision of professor
    Nirupam Chakraborti.

    For the MATLAB implementation, see:
    N. Chakraborti. Data-Driven Bi-Objective Genetic Algorithms EvoNN and BioGP and
    Their Applications in Metallurgical and Materials Domain. In Datta, Shubhabrata,
    Davim, J. Paulo (eds.), Computational Approaches to Materials Design: Theoretical
    and Practical Aspects, pp. 346-369, 2016.

    References
    ----------
    [1] Laumanns, M., Rudolph, G., & Schwefel, H. P. (1998). A spatial predator-prey
    approach to multi-objective
    optimization: A preliminary study.
    In International Conference on Parallel Problem Solving from Nature (pp. 241-249).
    Springer, Berlin, Heidelberg.

    [2] Li, X. (2003). A real-coded predator-prey genetic algorithm for multiobjective
    optimization. In International
    Conference on Evolutionary Multi-Criterion Optimization (pp. 207-221). Springer,
    Berlin, Heidelberg.

    [3] Chakraborti, N. (2014). Strategies for evolutionary data driven modeling in
    chemical and metallurgical Systems.
    In Applications of Metaheuristics in Process Engineering (pp. 89-122). Springer,
    Cham.

    [4] Pettersson, F., Chakraborti, N., & Saxén, H. (2007). A genetic algorithms based
    multi-objective neural net
    applied to noisy blast furnace data. Applied Soft Computing, 7(1), 387-397.

    """

    def __init__(
        self,
        problem,
        population_size: int = 100,
        population_params=None,
        initial_population=None,
        n_iterations: int = 10,
        n_gen_per_iter: int = 10,
        predator_pop_size: int = 50,
        prey_max_moves: int = 10,
        prob_prey_move: float = 0.3,
        offspring_place_attempts: int = 10,
        kill_interval: int = 7,
        max_rank: int = 20,
        neighbourhood_radius: int = 3,
    ):
        super().__init__(n_gen_per_iter=n_gen_per_iter, n_iterations=n_iterations)
        if initial_population is None:
            msg = "Provide initial population"
            raise eaError(msg)
        self.population = initial_population
        self.target_pop_size = population_size
        self.predator_pop_size: int = predator_pop_size
        self.prey_max_moves: int = prey_max_moves
        self.prob_prey_move: float = prob_prey_move
        self.offspring_place_attempts: int = offspring_place_attempts
        self.kill_interval: int = kill_interval
        self.max_rank: int = max_rank
        self.neighbourhood_radius: int = neighbourhood_radius
        self.lattice = Lattice(
            size_x=60,
            size_y=60,
            population=self.population,
            predator_pop_size=predator_pop_size,
            target_pop_size=self.target_pop_size,
            prob_prey_move=prob_prey_move,
            prey_max_moves=prey_max_moves,
            offspring_place_attempts=offspring_place_attempts,
            neighbourhood_radius=neighbourhood_radius,
        )

    def _next_gen(self):
        """Run one generation of PPGA.

        Intended to be used by next_iteration.

        Parameters
        ----------
        population: "Population"
            Population object
        """

        # Move prey and select neighbours for breeding
        mating_pop = self.lattice.move_prey()

        offspring = self.population.mate(mating_pop)

        # Try to place the offspring to lattice, add to population if successful
        placed_indices = self.lattice.place_offspring(len(offspring))

        # Remove from offsprings the ones that didn't get placed
        mask = np.ones(len(offspring), dtype=bool)
        mask[placed_indices] = False
        offspring = np.asarray(offspring)[~mask]

        # Add the successfully placed offspring to the population
        self.population.add(offspring)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1
        self._function_evaluation_count += offspring.shape[0]

        # Kill bad individuals every n generations
        if self._current_gen_count % self.kill_interval == 0:
            selected = self.select(self.population, self.max_rank)
            self.lattice.update_lattice(selected)
            self.population.delete(selected)

        # Move predators
        self.lattice.move_predator()

    def select(self, population, max_rank=20) -> list:
        """Of the population, individuals lower than max_rank are selected.
        Return indices of selected individuals.

        Parameters
        ----------
        population : Population
            Contains the current population and problem
            information.
        max_rank : int
            Select only individuals lower than max_rank

        Returns
        -------
        list
            List of indices of individuals to be selected.
        """
        # Calculating fronts and ranks
        # _, _, _, rank = nds(population.fitness)
        fronts = nds(population.fitness)
        index_ranks = np.argwhere(fronts)
        rank = np.full(fronts.shape[1], np.inf, dtype=int)
        rank[index_ranks[:, 1]] = index_ranks[:, 0]
        selection = np.nonzero(rank > max_rank)
        return selection[0]

    def manage_preferences(self, preference=None):
        return


class Lattice:
    """The 2-dimensional toroidal lattice in which the predators and prey are placed.

    Attributes
    ----------
    size_x : int
        Width of the lattice.
    size_y : int
        Height of the lattice.
    lattice : ndarray
        2d array for the lattice.
    predator_pop : ndarray
        The predator population.
    predators_loc : list
        Location (x, y) of predators on the lattice.
    preys_loc : list
        Location (x, y) of preys on the lattice.

    """

    def __init__(
        self,
        size_x,
        size_y,
        population,
        predator_pop_size,
        target_pop_size,
        prob_prey_move,
        prey_max_moves,
        offspring_place_attempts,
        neighbourhood_radius,
    ):

        self.size_x = size_x
        self.size_y = size_y
        self.population = population
        self.predator_pop_size = predator_pop_size
        self.target_pop_size = target_pop_size
        self.prob_prey_move = prob_prey_move
        self.prey_max_moves = prey_max_moves
        self.offspring_place_attempts = offspring_place_attempts
        self.neighbourhood_radius = neighbourhood_radius

        self.lattice = np.zeros((self.size_y, self.size_x), int)
        self.predator_pop = np.empty((0, 1))
        self.predators_loc = []
        self.preys_loc = []
        self.mating_pop = []
        self.init_predators()
        self.init_prey()

    def init_predators(self):
        """Initialize the predator population, linearly distributed in [0,1]
        and place them in the lattice randomly."""

        # Initialize the predator population
        self.predator_pop = np.linspace(0, 1, num=self.predator_pop_size)

        # Take random indices from free (==zero) lattice spaces
        free_space = np.transpose(np.nonzero(self.lattice == 0))
        indices = sample(free_space.tolist(), self.predator_pop.shape[0])

        for i in range(self.predator_pop.shape[0]):

            # Keep track of predator locations in a list
            self.predators_loc.append([indices[i][0], indices[i][1]])

            # +1 to offset zero index individual, set negative number to
            # identify predators from prey in the lattice
            self.lattice[indices[i][0]][indices[i][1]] = int(-1 * (i + 1))

    def init_prey(self):
        """Find an empty position in the lattice and place the prey."""

        # Take random indices from free (==zero) lattice spaces
        free_space = np.transpose(np.nonzero(self.lattice == 0))
        indices = sample(free_space.tolist(), len(self.population.individuals))

        for i in range(len(self.population.individuals)):

            # Keep track of preys in a list
            self.preys_loc.append([indices[i][0], indices[i][1]])

            # +1 to offset zero index individual
            self.lattice[indices[i][0]][indices[i][1]] = int(i + 1)

    def move_prey(self):
        """Find an empty position in prey neighbourhood for the prey to move in,
        and choose a mate for breeding if any available.

        Returns
        -------
        mating_pop : list
            List of parent indices to use for mating
        """
        mating_pop = []
        for prey, pos in enumerate(self.preys_loc):

            if np.random.random() < self.prob_prey_move:

                for i in range(self.prey_max_moves):

                    neighbours = self.neighbours(self.lattice, pos[0], pos[1])

                    dy = np.random.randint(neighbours.shape[0])
                    dx = np.random.randint(neighbours.shape[1])

                    # If neighbouring cell is occupied, skip turn
                    if neighbours[dy][dx] != 0:
                        continue

                    dest_y = dy - 1 + pos[0]
                    dest_x = dx - 1 + pos[1]

                    # Check boundaries of the lattice
                    if dest_y not in range(self.size_y) or dest_x not in range(
                        self.size_x
                    ):
                        dest_y, dest_x = self.lattice_wrap_idx(
                            (dest_y, dest_x), np.shape(self.lattice)
                        )

                    # Move prey, clear previous location
                    self.lattice[dest_y][dest_x] = int(prey + 1)
                    self.lattice[pos[0]][pos[1]] = 0

                    # Update prey location in the list
                    pos[0], pos[1] = dest_y, dest_x

            neighbours = self.neighbours(
                self.lattice, self.preys_loc[prey][0], self.preys_loc[prey][1]
            )
            mates = neighbours[(neighbours > 0) & (neighbours != prey + 1)]
            if len(mates) < 1:
                continue
            else:
                # -1 for lattice offset
                mate = int(choice(mates)) - 1
                mating_pop.append([prey, mate])
        if mating_pop == []:
            raise eaError("What's ahppening?!")
        return mating_pop

    def place_offspring(self, offspring):
        """Try to place the offsprings to the lattice. If no empty spot found within
        number of max attempts, do not place.

        Parameters
        ----------
        offspring : int
            number of offsprings

        Returns
        -------
        list
            Successfully placed offspring indices.
        """

        # Keep track of offspring in a list
        placed_offspring = []

        for i in range(offspring):

            y = np.random.randint(self.size_y)
            x = np.random.randint(self.size_x)

            for j in range(self.offspring_place_attempts):
                if self.lattice[y][x] != 0:
                    continue

                else:

                    if self.lattice[y][x] == 0:
                        # Append the offspring to the list of preys.
                        # len(self.preys_loc) is the index of the current
                        # last prey in the list
                        self.lattice[y][x] = int(len(self.preys_loc) + 1)
                        self.preys_loc.append([y, x])
                        placed_offspring.append(i)

        return placed_offspring

    def move_predator(self):
        """Find an empty position in the predator neighbourhood for the predators to move in,
        move the predator and kill the weakest prey in its neighbourhood, if any.
        Repeat until > predator_max_moves."""

        predator_max_moves = int(
            (len(self.population.individuals) - self.target_pop_size)
            / self.predator_pop_size
        )

        # Track killed preys in list and remove them at the end of the function
        to_be_killed = []

        for predator, pos in enumerate(self.predators_loc):

            for i in range(predator_max_moves):

                neighbours = self.neighbours(
                    self.lattice, pos[0], pos[1], n=self.neighbourhood_radius
                )
                targets = neighbours[neighbours > 0]

                # If preys found in the neighbourhood,
                # calculate their fitness and kill the weakest
                if len(targets) > 0:
                    fitness = []
                    weakest_prey = None
                    for target in targets:

                        obj1 = self.population.fitness[target - 1][0]
                        obj2 = self.population.fitness[target - 1][1]

                        fc = (
                            self.predator_pop[predator] * obj1
                            + (1 - self.predator_pop[predator]) * obj2
                        )

                        fitness.append((fc, target))
                        fitness.sort()
                        weakest_prey = fitness[-1][1] - 1

                    # Kill the weakest prey and move the predator to its place
                    self.lattice[self.preys_loc[weakest_prey][0]][
                        self.preys_loc[weakest_prey][1]
                    ] = -1 * (predator + 1)

                    # Set the old predator location to zero
                    self.lattice[pos[0]][pos[1]] = 0

                    # Update predator location in the list of predators
                    pos[0], pos[1] = (
                        self.preys_loc[weakest_prey][0],
                        self.preys_loc[weakest_prey][1],
                    )

                    # Set the killed prey as dead in the list of prey locations and
                    # end predator turn
                    to_be_killed.append(weakest_prey)
                    self.preys_loc[weakest_prey] = None

                else:
                    dy = np.random.randint(neighbours.shape[0])
                    dx = np.random.randint(neighbours.shape[1])

                    # If neighbouring cell is occupied by another predator, skip turn
                    if neighbours[dy][dx] < 0:
                        continue

                    dest_y = dy - 1 + pos[0]
                    dest_x = dx - 1 + pos[1]

                    # Check boundaries of the lattice
                    if dest_y not in range(self.size_y) or dest_x not in range(
                        self.size_x
                    ):
                        dest_y, dest_x = self.lattice_wrap_idx(
                            (dest_y, dest_x), np.shape(self.lattice)
                        )

                    # Move predator, clear previous location
                    self.lattice[dest_y][dest_x] = -1 * (predator + 1)
                    self.lattice[pos[0]][pos[1]] = 0

                    # Update predator location in the list
                    pos[0], pos[1] = dest_y, dest_x

        # Remove killed prey from population
        self.population.delete(to_be_killed)
        self.update_lattice()

    def update_lattice(self, selected=None):
        """Update prey positions in the lattice.

        Parameters
        ----------
        selected : list
            Indices of preys to be removed from the lattice.

        """
        # Remove selected individuals from the lattice
        if selected is not None:
            for i in selected:
                self.lattice[self.preys_loc[i][0]][self.preys_loc[i][1]] = 0
                self.preys_loc[i] = None

        # Update the list of prey locations
        updated_preys = [x for x in self.preys_loc if x is not None]
        self.preys_loc = updated_preys

        # Update lattice
        for prey, pos in enumerate(self.preys_loc):
            self.lattice[pos[0]][pos[1]] = prey + 1

    @staticmethod
    def lattice_wrap_idx(index, lattice_shape):
        """Returns periodic lattice index
        for a given iterable index.

        Parameters
        ----------
        index : tuple
            one integer for each axis
        lattice_shape : tuple
            the shape of the lattice to index to
        """

        if not hasattr(index, "__iter__"):
            return index  # handle integer slices
        if len(index) != len(lattice_shape):
            return index  # must reference a scalar
        if any(type(i) == slice for i in index):
            return index  # slices not supported
        if len(index) == len(lattice_shape):  # periodic indexing of scalars
            mod_index = tuple(((i % s + s) % s for i, s in zip(index, lattice_shape)))
            return mod_index
        raise ValueError("Unexpected index: {}".format(index))

    @staticmethod
    def neighbours(arr, x, y, n=3):
        """Given a 2D-array, returns an n*n array whose "center" element is arr[x,y]

        Parameters
        ----------
        arr : ndarray
            A 2D-array where to get the neighbouring cells
        x : int
            X coordinate for the center element
        y : int
            Y coordinate for the center element
        n : int
            Radius of the neighbourhood

        Returns
        -------
        The neighbouring cells of x, y in radius n*n.
        Defaults to Moore neighbourhood (n=3).
        """

        arr = np.roll(np.roll(arr, shift=-x + 1, axis=0), shift=-y + 1, axis=1)
        return arr[:n, :n]



import numpy as np
import random
from math import ceil
from pyDOE import lhs


def create_new_individuals(design, problem, pop_size=None):
    """Create new individuals to the population.

    The individuals can be created randomly, by LHS design, or can be passed by the
    user.

    Design does not apply in case of EvoNN and EvoDN2 problem, where neural networks
    are created as individuals.

    Parameters
    ----------
    design : str, optional
        Describe the method of creation of new individuals.
        "RandomDesign" creates individuals randomly.
        "LHSDesign" creates individuals using Latin hypercube sampling.
        "EvoNN" creates Artificial Neural Networks as individuals.
        "EvoDN2" creates Deep Neural Networks.
    problem : baseProblem
        An object of the class Problem
    pop_size : int, optional
        Number of individuals in the population. If none, some default population
        size based on number of objectives is chosen.

    Returns
    -------
    individuals : list
        A list of individuals.

    """

    if pop_size is None:
        pop_size_options = [50, 105, 120, 126, 132, 112, 156, 90, 275]
        pop_size = pop_size_options[problem.num_of_objectives - 2]

    if design == "RandomDesign":
        lower_limits = np.asarray(problem.get_variable_lower_bounds())
        upper_limits = np.asarray(problem.get_variable_upper_bounds)
        individuals = np.random.random((pop_size, problem.n_of_variables))
        # Scaling
        individuals = individuals * (upper_limits - lower_limits) + lower_limits

        return individuals

    elif design == "LHSDesign":
        lower_limits = np.asarray(problem.get_variable_lower_bounds())
        upper_limits = np.asarray(problem.get_variable_upper_bounds())
        individuals = lhs(problem.n_of_variables, samples=pop_size)
        # Scaling
        individuals = individuals * (upper_limits - lower_limits) + lower_limits

        return individuals

    elif design == "EvoNN":

        """Create a population of neural networks for the EvoNN algorithm.

        Individuals are 2d arrays representing the weight matrices of the NNs.
        One extra row is added for bias.

        """

        w_low = problem.params["w_low"]
        w_high = problem.params["w_high"]
        in_nodes = problem.num_of_variables
        num_nodes = problem.params["num_nodes"]
        prob_omit = problem.params["prob_omit"]

        individuals = np.random.uniform(
            w_low, w_high, size=(pop_size, in_nodes, num_nodes)
        )

        # Randomly set some weights to zero
        zeros = np.random.choice(
            np.arange(individuals.size), ceil(individuals.size * prob_omit)
        )
        individuals.ravel()[zeros] = 0

        # Set bias
        individuals = np.insert(individuals, 0, 1, axis=1)

        return individuals

    elif design == "EvoDN2":
        """Create a population of deep neural networks (DNNs) for the EvoDN2 algorithm.

        Each individual is a list of subnets, and each subnet contains a random amount
        of layers and
        nodes per layer. The subnets are evolved via evolutionary algorithms, and they
        converge
        on the final linear layer of the DNN.
        """

        individuals = []
        for i in range(problem.params["pop_size"]):
            nets = []
            for j in range(problem.params["num_subnets"]):

                layers = []
                num_layers = np.random.randint(1, problem.params["max_layers"])
                in_nodes = len(problem.subsets[j])

                for k in range(num_layers):
                    out_nodes = random.randint(2, problem.params["max_nodes"])
                    net = np.random.uniform(
                        problem.params["w_low"],
                        problem.params["w_high"],
                        size=(in_nodes, out_nodes),
                    )
                    # Randomly set some weights to zero
                    zeros = np.random.choice(
                        np.arange(net.size),
                        ceil(net.size * problem.params["prob_omit"]),
                    )
                    net.ravel()[zeros] = 0

                    # Add bias
                    net = np.insert(net, 0, 1, axis=0)
                    in_nodes = out_nodes
                    layers.append(net)

                nets.append(layers)

            individuals.append(nets)

        return individuals

    elif design == "BioGP":
        return problem.create_individuals()



import numpy as np
from copy import deepcopy
from random import choice
from random import shuffle


def mate(mating_pop, individuals: list, params):
    """Perform BioGP crossover functions. Produce two offsprings by swapping genetic
    material of the two parents.

    Standard crossover:
    Swap two random subtrees between the parents.

    Height-fair crossover:
    Swap two random subtrees between the parents at the selected depth.

    Parameters
    ----------
    mating_pop : list
        List of indices of individuals to mate. If None, choose from population
        randomly.
        Each entry should contain two indices, one for each parent.
    individuals : list
        List of all individuals.
    params : dict
        Parameters for evolution. If None, use defaults.

    Returns
    -------
    offspring : list
        The offsprings produced as a result of crossover.
    """

    prob_crossover = params.get("prob_crossover", 0.9)

    prob_standard = 0.5
    prob_height_fair = prob_crossover - prob_standard
    r = np.random.rand()

    if mating_pop is None:
        mating_pop = []
        for i in range(len(individuals)):
            mating_pop.append([i, np.random.randint(len(individuals))])

    offspring = []

    for mates in mating_pop:

        offspring1 = deepcopy(individuals[mates[0]])
        offspring2 = deepcopy(individuals[mates[1]])

        # Height-fair xover
        if r <= prob_height_fair:
            depth = min(offspring1.total_depth, offspring2.total_depth)
            rand_node1 = choice(offspring1.nodes_at_depth[depth])
            rand_node2 = choice(offspring2.nodes_at_depth[depth])
            tmp_value = rand_node1.value
            tmp_roots = rand_node1.roots
            rand_node1.value = rand_node2.value
            rand_node1.roots = rand_node2.roots
            rand_node2.value = tmp_value
            rand_node2.roots = tmp_roots
            offspring1.nodes = offspring1.get_sub_nodes()
            offspring2.nodes = offspring2.get_sub_nodes()

        # Standard xover
        elif r <= prob_height_fair + prob_standard:
            rand_node1 = choice(offspring1.nodes[1:])  # Exclude linear node
            rand_node2 = choice(offspring2.nodes[1:])
            tmp_value = rand_node1.value
            tmp_roots = rand_node1.roots
            rand_node1.value = rand_node2.value
            rand_node1.roots = rand_node2.roots
            rand_node2.value = tmp_value
            rand_node2.roots = tmp_roots
            offspring1.nodes = offspring1.get_sub_nodes()
            offspring2.nodes = offspring2.get_sub_nodes()

        else:
            pass

        offspring.extend((offspring1, offspring2))

    return offspring


class BioGP_xover:
    def __init__(
        self, probability_crossover: float = 0.9, probability_standard: float = 0.5
    ):
        self.probability_crossover: float = probability_crossover
        if probability_standard > probability_crossover:
            msg = (
                f"Probability of standard crossover ({probability_standard}) should be "
                f"less than probability of crossover ({probability_crossover})"
            )
            raise ValueError(msg)
        self.probability_standard: float = probability_standard

    def do(self, pop, mating_pop_ids):
        prob_height_fair = self.probability_crossover - self.probability_standard
        pop_size = pop.shape[0]
        if mating_pop_ids is None:
            shuffled_ids = list(range(pop_size))
            shuffle(shuffled_ids)
        else:
            shuffled_ids = mating_pop_ids
        if np.asarray(shuffled_ids).ndim == 1:
            mating_pop = [
                [shuffled_ids[x], shuffled_ids[x]]
                for x in range(len(shuffled_ids))
                if x % 2 == 0
            ]
        elif np.asarray(shuffled_ids).ndim == 2:
            mating_pop = shuffled_ids
        offspring = []
        r = np.random.rand()
        for mates in mating_pop:

            offspring1 = deepcopy(pop[mates[0]][0])
            offspring2 = deepcopy(pop[mates[1]][0])

            # Height-fair xover
            if r <= prob_height_fair:
                depth = min(offspring1.total_depth, offspring2.total_depth)
                rand_node1 = choice(offspring1.nodes_at_depth[depth])
                rand_node2 = choice(offspring2.nodes_at_depth[depth])
                tmp_value = rand_node1.value
                tmp_roots = rand_node1.roots
                rand_node1.value = rand_node2.value
                rand_node1.roots = rand_node2.roots
                rand_node2.value = tmp_value
                rand_node2.roots = tmp_roots
                offspring1.nodes = offspring1.get_sub_nodes()
                offspring2.nodes = offspring2.get_sub_nodes()

            # Standard xover
            elif r <= self.probability_crossover:
                rand_node1 = choice(offspring1.nodes[1:])  # Exclude linear node
                rand_node2 = choice(offspring2.nodes[1:])
                tmp_value = rand_node1.value
                tmp_roots = rand_node1.roots
                rand_node1.value = rand_node2.value
                rand_node1.roots = rand_node2.roots
                rand_node2.value = tmp_value
                rand_node2.roots = tmp_roots
                offspring1.nodes = offspring1.get_sub_nodes()
                offspring2.nodes = offspring2.get_sub_nodes()

            else:
                pass

            offspring.extend((offspring1, offspring2))

        return np.asarray(offspring).reshape(-1,1)



import numpy as np


class BP_mutation:
    def __init__(
        self,
        lower_limits: np.ndarray,
        upper_limits: np.ndarray,
        ProM: float = None,
        DisM: float = 20,
    ):
        self.lower_limits = lower_limits
        self.upper_limits = upper_limits
        if ProM is None:
            self.ProM = 1/len(lower_limits)
        else:
            self.ProM = ProM
        self.DisM = DisM

    def do(self, offspring: np.ndarray):
        """Conduct bounded polynomial mutation. Return the mutated individuals.

        Parameters
        ----------
        offspring : np.ndarray
            The array of offsprings to be mutated.

        Returns
        -------
        np.ndarray
            The mutated offsprings
        """
        min_val = np.ones_like(offspring) * self.lower_limits
        max_val = np.ones_like(offspring) * self.upper_limits
        k = np.random.random(offspring.shape)
        miu = np.random.random(offspring.shape)
        temp = np.logical_and((k <= self.ProM), (miu < 0.5))
        offspring_scaled = (offspring - min_val) / (max_val - min_val)
        offspring[temp] = offspring[temp] + (
            (max_val[temp] - min_val[temp])
            * (
                (
                    2 * miu[temp]
                    + (1 - 2 * miu[temp])
                    * (1 - offspring_scaled[temp]) ** (self.DisM + 1)
                )
                ** (1 / (self.DisM + 1))
                - 1
            )
        )
        temp = np.logical_and((k <= self.ProM), (miu >= 0.5))
        offspring[temp] = offspring[temp] + (
            (max_val[temp] - min_val[temp])
            * (
                1
                - (
                    2 * (1 - miu[temp])
                    + 2 * (miu[temp] - 0.5) * offspring_scaled[temp] ** (self.DisM + 1)
                )
                ** (1 / (self.DisM + 1))
            )
        )
        offspring[offspring > max_val] = max_val[offspring > max_val]
        offspring[offspring < min_val] = min_val[offspring < min_val]
        return offspring



import numpy as np
from typing import List
from desdeo_emo.selection.SelectionBase import SelectionBase
from desdeo_emo.population.Population import Population
from desdeo_emo.othertools.ReferenceVectors import ReferenceVectors
from desdeo_tools.scalarization.MOEADSF import (
    MOEADSFBase,
    Tchebycheff,
    PBI,
    WeightedSum,
)


class MOEAD_select(SelectionBase):
    """The MOEAD selection operator. 

    Parameters
    ----------
    pop : Population
        The population of individuals
    SF_type : MOEADSFBase
        The scalarizing function employed to evaluate the solutions

    """

    def __init__(self, pop: Population, SF_type: MOEADSFBase):
        # initialize
        self.SF_type = SF_type

    def do(
        self,
        pop: Population,
        vectors: ReferenceVectors,
        ideal_point,
        current_neighborhood,
        offspring_fx,
    ) -> List[int]:
        """Select the individuals that are kept in the neighborhood.

        Parameters
        ----------
        pop : Population
            The current population.
        vectors : ReferenceVectors
            Class instance containing reference vectors.
        ideal_point
            Ideal vector found so far
        current_neighborhood
            Neighborhood to be updated
        offspring_fx
            Offspring solution to be compared with the rest of the neighborhood

        Returns
        -------
        List[int]
            List of indices of the selected individuals
        """
        # Compute the value of the SF for each neighbor
        num_neighbors = len(current_neighborhood)
        current_population = pop.objectives[current_neighborhood, :]
        current_reference_vectors = vectors.values[current_neighborhood, :]
        offspring_population = np.tile(offspring_fx, (num_neighbors, 1))
        ideal_point_matrix = np.tile(ideal_point, (num_neighbors, 1))

        values_SF = self._evaluate_SF(
            current_population, current_reference_vectors, ideal_point_matrix
        )
        values_SF_offspring = self._evaluate_SF(
            offspring_population, current_reference_vectors, ideal_point_matrix
        )

        # Compare the offspring with the individuals in the neighborhood
        # and replace the ones which are outperformed by it.
        selection = np.where(values_SF_offspring < values_SF)[0]

        return current_neighborhood[selection]

    def _evaluate_SF(self, neighborhood, weights, ideal_point):
        SF_values = np.array(
            list(map(self.SF_type, neighborhood, weights, ideal_point))
        )
        return SF_values




from abc import ABC, abstractmethod
from typing import Dict, List, Union

import numpy as np

from desdeo_emo.population.CreateIndividuals import create_new_individuals
from desdeo_emo.recombination.BoundedPolynomialMutation import BP_mutation
from desdeo_emo.recombination.SimulatedBinaryCrossover import SBX_xover
from desdeo_problem.Problem import MOProblem


class BasePopulation(ABC):
    def __init__(self, problem: MOProblem, pop_size: int, pop_params: Dict = None):
        self.pop_size: int = pop_size
        self.problem = problem
        self.individuals: Union[List, np.ndarray] = None
        self.objectives: np.ndarray = None
        self.uncertainity: np.ndarray = None
        self.fitness: np.ndarray = None
        if not problem.n_of_constraints == 0:
            self.constraint = None
        self.ideal_objective_vector = problem.ideal
        self.nadir_objective_vector = problem.nadir
        self.ideal_fitness_val = None
        self.nadir_fitness_val = None
        if problem.ideal is not None:
            self.ideal_fitness_val = problem.ideal * problem._max_multiplier
            self.nadir_fitness_val = problem.nadir * problem._max_multiplier
        self.xover = None
        self.mutation = None
        self.recombination = None

    @abstractmethod
    def add(self, offsprings: Union[List, np.ndarray]) -> List:
        """Evaluate and add offspring to the population.

        Parameters
        ----------
        offsprings : Union[List, np.ndarray]
            List or array of individuals to be evaluated and added to the population.

        Returns
        -------
        List
            Indices of the evaluated individuals
        """
        pass

    @abstractmethod
    def keep(self, indices: List):
        """Save the population members given by the list of indices for the next
            generation. Delete the rest.

        Parameters
        ----------
        indices : List
            List of indices of the population members to be kept for the next
                generation.
        """
        pass

    @abstractmethod
    def delete(self, indices: List):
        """Delete the population members given by the list of indices for the next
            generation. Keep the rest.

        Parameters
        ----------
        indices : List
            List of indices of the population members to be deleted.
        """
        pass

    @abstractmethod
    def mate(
        self, mating_individuals: List = None, params: Dict = None
    ) -> Union[List, np.ndarray]:
        """Perform crossover and mutation over the population members.

        Parameters
        ----------
        mating_individuals : List, optional
            List of individuals taking part in recombination. By default None, which
                recombinated all individuals in random order.
        params : Dict, optional
            Parameters for the mutation or crossover operator, by default None.

        Returns
        -------
        Union[List, np.ndarray]
            The offspring population
        """
        pass


class Population(BasePopulation):
    def __init__(
        self,
        problem: MOProblem,
        pop_size: int,
        pop_params: Dict = None,
        use_surrogates: bool = False,
    ):
        super().__init__(problem, pop_size)
        self.lower_limits = self.problem.get_variable_lower_bounds()
        self.upper_limits = self.problem.get_variable_upper_bounds()
        if pop_params is None:
            design = "LHSDesign"
        if pop_params is not None:
            if "design" in pop_params.keys():
                design = pop_params["design"]
            else:
                design = "LHSDesign"
        individuals = create_new_individuals(design, problem, pop_size)
        self.add(individuals, use_surrogates)
        self.xover = SBX_xover()
        self.mutation = BP_mutation(self.lower_limits, self.upper_limits)

    def add(
        self, offsprings: Union[List, np.ndarray], use_surrogates: bool = False
    ) -> List:
        """Evaluate and add offspring to the population.

        Parameters
        ----------
        offsprings : Union[List, np.ndarray]
            List or array of individuals to be evaluated and added to the population.
        
        use_surrogates: bool
            If true, use surrogate models rather than true function evaluations.

        use_surrogates: bool
            If true, use surrogate models rather than true function evaluations.

        Returns
        -------
        List
            Indices of the evaluated individuals
        """
        results = self.problem.evaluate(offsprings, use_surrogates)
        objectives = results.objectives
        fitness = results.fitness
        constraints = results.constraints
        uncertainity = results.uncertainity
        if self.individuals is None:
            self.individuals = offsprings
            self.objectives = objectives
            self.fitness = fitness
            self.constraint = constraints
            self.uncertainity = uncertainity
            first_offspring_index = 0
        else:
            first_offspring_index = self.individuals.shape[0]
            if self.individuals.ndim - offsprings.ndim == 1:
                self.individuals = np.vstack((self.individuals, [offsprings]))
            elif self.individuals.ndim == offsprings.ndim:
                self.individuals = np.vstack((self.individuals, offsprings))
            else:
                pass  # TODO raise error
            self.objectives = np.vstack((self.objectives, objectives))
            self.fitness = np.vstack((self.fitness, fitness))
            if self.problem.n_of_constraints != 0:
                self.constraint = np.vstack(
                    (self.constraint, constraints)
                )
            if uncertainity is None:
                self.uncertainity = None
            else:    
                self.uncertainity = np.vstack((self.uncertainity, uncertainity))
        last_offspring_index = self.individuals.shape[0]
        self.update_ideal()
        return list(range(first_offspring_index, last_offspring_index))

    def keep(self, indices: List):
        """Save the population members given by the list of indices for the next
            generation. Delete the rest.

        Parameters
        ----------
        indices : List
            List of indices of the population members to be kept for the next
                generation.
        """
        mask = np.zeros(self.individuals.shape[0], dtype=bool)
        mask[indices] = True
        self.individuals = self.individuals[mask]
        self.objectives = self.objectives[mask]
        self.fitness = self.fitness[mask]
        if self.uncertainity is not None:
            self.uncertainity = self.uncertainity[mask]
        if self.problem.n_of_constraints > 0:
            self.constraint = self.constraint[mask]

    def delete(self, indices: List):
        """Delete the population members given by the list of indices for the next
            generation. Keep the rest.

        Parameters
        ----------
        indices : List
            List of indices of the population members to be deleted.
        """
        mask = np.ones(self.individuals.shape[0], dtype=bool)
        mask[indices] = False
        self.individuals = self.individuals[mask]
        if len(self.individuals) == 0:
            self.individuals = None
        self.objectives = self.objectives[mask]
        self.fitness = self.fitness[mask]
        if self.uncertainity is not None:
            self.uncertainity = self.uncertainity[mask]
        if self.problem.n_of_constraints > 0:
            self.constraint = self.constraint[mask]

    def mate(self, mating_individuals: List = None) -> Union[List, np.ndarray]:
        """Perform crossover and mutation over the population members.

        Parameters
        ----------
        mating_individuals : List, optional
            List of individuals taking part in recombination. By default None, which
                recombinated all individuals in random order.
        params : Dict, optional
            Parameters for the mutation or crossover operator, by default None.

        Returns
        -------
        Union[List, np.ndarray]
            The offspring population
        """
        if self.recombination is not None:
            offspring = self.recombination.do(self.individuals, mating_individuals)
        else:
            offspring = self.xover.do(self.individuals, mating_individuals)
            offspring = self.mutation.do(offspring)
        return offspring

    def update_ideal(self):
        if self.ideal_fitness_val is None:
            self.ideal_fitness_val = np.amin(self.fitness, axis=0)
        else:
            self.ideal_fitness_val = np.amin(
                np.vstack((self.ideal_fitness_val, self.fitness)), axis=0
            )
        self.ideal_objective_vector = (
            self.ideal_fitness_val * self.problem._max_multiplier
        )
    def replace(self, indices: List, individual:np.ndarray, evaluation:tuple):
        """Replace the population members given by the list of indices by the given individual and its evaluation. 
           Keep the rest of the population unchanged.

        Parameters
        ----------
        indices : List
            List of indices of the population members to be replaced.
        individual: np.ndarray
            Decision variables of the individual that will replace the positions given in the list.
        evaluation: tuple
            Result of the evaluation of the objective function, constraints, etc. obtained using the evaluate method.
        """
        self.individuals[indices, :] = individual
        self.objectives[indices, :] = evaluation.objectives
        self.fitness[indices, :] = evaluation.fitness
        if self.constraint is not None:
            self.constraint[indices, :] = evaluation.constraints
        if self.uncertainity is not None:
            self.uncertainity[indices, :] = evaluation.uncertainity

    def repair(self, individual):
        """Repair the variables of an individual which are not in the boundary defined by the problem
        Parameters
        ----------
        individual : 
            Decision variables of the individual.

        Return
        ----------
        The new decision vector with the variables in the boundary defined by the problem
        """
        upper_bounds = self.problem.get_variable_upper_bounds()
        lower_bounds = self.problem.get_variable_lower_bounds()
        upper_bounds_check = np.where(individual > upper_bounds)
        lower_bounds_check = np.where(individual < lower_bounds)
        individual[upper_bounds_check] = upper_bounds[upper_bounds_check]
        individual[lower_bounds_check] = lower_bounds[lower_bounds_check]
        return individual


"""This module contains classes implementing different Evolutionary algorithms.
"""

__all__ = [
    "RVEA",
    "NSGAIII",
    "BaseEA",
    "BaseDecompositionEA",
    "PPGA",
    "TournamentEA",
    "IOPIS_NSGAIII",
    "IOPIS_RVEA",
]

from desdeo_emo.EAs.BaseEA import BaseEA, BaseDecompositionEA
from desdeo_emo.EAs.RVEA import RVEA
from desdeo_emo.EAs.NSGAIII import NSGAIII
from desdeo_emo.EAs.PPGA import PPGA
from desdeo_emo.EAs.TournamentEA import TournamentEA
from desdeo_emo.EAs.IOPIS import IOPIS_NSGAIII, IOPIS_RVEA


from abc import ABC, abstractmethod
import numpy as np
from typing import List


class SelectionBase(ABC):
    """The base class for the selection operator.
    """
    @abstractmethod
    def do(self, fitness: np.ndarray, *args) -> List[int]:
        """Use the selection operator over the given fitness values. Return the indices
            individuals with the best fitness values according to the operator.

        Parameters
        ----------
        fitness : np.ndarray
            Fitness of the individuals from which the next generation is to be selected.

        Returns
        -------
        List[int]
            The list of selected individuals
        """
        pass



"""This module provides implementations of various selection operators.
"""

__all__ = ["APD_Select", "NSGAIII_select", "tour_select"]

from desdeo_emo.selection.APD_Select_constraints import APD_Select
from desdeo_emo.selection.NSGAIII_select import NSGAIII_select
from desdeo_emo.selection.tournament_select import tour_select



"""This module provides classes and methods which implement populations in an EA."""

__all__ = ["Population", "create_new_individuals", "SurrogatePopulation"]

from desdeo_emo.population.CreateIndividuals import create_new_individuals
from desdeo_emo.population.Population import Population
from desdeo_emo.population.SurrogatePopulation import SurrogatePopulation



"""This module provides miscellaneous tools"""

__all__ = ["IsNotebook", "animate_init_", "animate_next_", "ReferenceVectors"]

from desdeo_emo.othertools.IsNotebook import IsNotebook
from desdeo_emo.othertools.plotlyanimate import animate_init_, animate_next_
from desdeo_emo.othertools.ReferenceVectors import ReferenceVectors



from desdeo_emo.selection.NSGAIII_select import NSGAIII_select
import numpy as np
#from pygmo import fast_non_dominated_sorting as nds
from desdeo_tools.utilities import fast_non_dominated_sort
from typing import List
from desdeo_emo.population.Population import Population
from desdeo_emo.othertools.ReferenceVectors import ReferenceVectors


class IOPIS_NSGAIII_select(NSGAIII_select):
    def __init__(
        self,
        scalarization_nethods,
        pop: Population,
        n_survive: int = None,
        selection_type: str = None,
    ):
        self.scalarization_methods = scalarization_nethods
        self.worst_fitness: np.ndarray = -np.full(
            (1, len(scalarization_nethods)), np.inf
        )
        self.extreme_points: np.ndarray = None
        if n_survive is None:
            self.n_survive: int = pop.pop_size
        self.ideal: np.ndarray = np.full((1, len(scalarization_nethods)), np.inf)

    def do(
        self, pop: Population, vectors: ReferenceVectors, reference_point: np.ndarray
    ) -> List[int]:
        """Select individuals for mating for NSGA-III.

        Parameters
        ----------
        pop : Population
            The current population.
        vectors : ReferenceVectors
            Class instance containing reference vectors.

        Returns
        -------
        List[int]
            List of indices of the selected individuals
        """
        ref_dirs = vectors.values_planar
        fitness = np.asarray(
            [
                scalar(pop.fitness, reference_point)
                for scalar in self.scalarization_methods
            ]
        ).T
        # Calculating fronts and ranks
        # fronts, dl, dc, rank = nds(fitness)
        fronts = fast_non_dominated_sort(fitness)
        fronts = [np.where(fronts[i])[0] for i in range(len(fronts))]
        non_dominated = fronts[0]
        fmin = np.amin(fitness, axis=0)
        self.ideal = np.amin(np.vstack((self.ideal, fmin)), axis=0)

        # Calculating worst points
        self.worst_fitness = np.amax(np.vstack((self.worst_fitness, fitness)), axis=0)
        worst_of_population = np.amax(fitness, axis=0)
        worst_of_front = np.max(fitness[non_dominated, :], axis=0)
        self.extreme_points = self.get_extreme_points_c(
            fitness[non_dominated, :], self.ideal, extreme_points=self.extreme_points
        )
        nadir_point = self.get_nadir_point(
            self.extreme_points,
            self.ideal,
            self.worst_fitness,
            worst_of_population,
            worst_of_front,
        )

        # Finding individuals in first 'n' fronts
        selection = np.asarray([], dtype=int)
        for front_id in range(len(fronts)):
            if len(np.concatenate(fronts[: front_id + 1])) < self.n_survive:
                continue
            else:
                fronts = fronts[: front_id + 1]
                selection = np.concatenate(fronts)
                break

        F = fitness[selection]

        last_front = fronts[-1]

        # Selecting individuals from the last acceptable front.
        if len(selection) > self.n_survive:
            niche_of_individuals, dist_to_niche = self.associate_to_niches(
                F, ref_dirs, self.ideal, nadir_point
            )
            # if there is only one front
            if len(fronts) == 1:
                n_remaining = self.n_survive
                until_last_front = np.array([], dtype=np.int)
                niche_count = np.zeros(len(ref_dirs), dtype=np.int)

            # if some individuals already survived
            else:
                until_last_front = np.concatenate(fronts[:-1])
                id_until_last_front = list(range(len(until_last_front)))
                niche_count = self.calc_niche_count(
                    len(ref_dirs), niche_of_individuals[id_until_last_front]
                )
                n_remaining = self.n_survive - len(until_last_front)

            last_front_selection_id = list(range(len(until_last_front), len(selection)))
            if np.any(selection[last_front_selection_id] != last_front):
                print("error!!!")
            selected_from_last_front = self.niching(
                fitness[last_front, :],
                n_remaining,
                niche_count,
                niche_of_individuals[last_front_selection_id],
                dist_to_niche[last_front_selection_id],
            )
            final_selection = np.concatenate(
                (until_last_front, last_front[selected_from_last_front])
            )
            if self.extreme_points is None:
                print("Error")
            if final_selection is None:
                print("Error")
        else:
            final_selection = selection
        return final_selection.astype(int)




"Legacy code. To be updated or depreciated."

from pyrvea.EAs.RVEA import RVEA
from pyrvea.OtherTools.ReferenceVectors import ReferenceVectors
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyrvea.Population.Population import Population


class slowRVEA(RVEA):
    """RVEA variant that impliments slow reference vector movement."""

    def __init__(self, population: "Population", ea_parameters):
        """Initialize a Base Decomposition EA.

        This will call methods to set up the parameters of RVEA, create
        Reference Vectors, and (as of Feb 2019) run the first iteration of RVEA.

        Parameters
        ----------
        population : "Population"
            This variable is updated as evolution takes place
        EA_parameters : dict
            Takes the EA parameters

        Returns
        -------
        Population:
            Returns the Population after evolution.
        """
        if ea_parameters:
            self.params = self.set_params(population, **ea_parameters)
        else:
            self.params = self.set_params(population)
        # if population.individuals.shape[0] == 0:
        #     create_new_individuals(pop_size=self.params["population_size"])
        # # print("Using BaseDecompositionEA init")
        # self._next_iteration(population)

    def set_params(
        self,
        population: "Population",
        generations_per_iteration: int = 10,
        iterations: int = 10,
        Alpha: float = 2,
        ref_point: list = None,
        old_point: list = None,
        **kwargs
    ):
        """Set up the parameters. Save in RVEA.params. Note, this should be
        changed to align with the current structure.

        Parameters
        ----------
        population : Population
            Population object
        Alpha : float
            The alpha parameter of APD selection.
        plotting : bool
            Useless really.
        Returns
        -------

        """
        ref_vectors = ReferenceVectors(
            number_of_objectives=population.problem.num_of_objectives,
            creation_type="Sparse_Focused",
            ref_point=old_point,
        )
        if ref_point is None:
            ref_point = ref_vectors.values[0]
        rveaparams = {
            "reference_vectors": ref_vectors,
            "population_size": ref_vectors.number_of_vectors,
            "generations": generations_per_iteration,
            "iterations": iterations,
            "Alpha": Alpha,
            "current_iteration_gen_count": 0,
            "current_iteration_count": 0,
            "current_total_gen_count": 0,
            "total_generations": iterations * generations_per_iteration,
            "ref_point": ref_point,
        }
        rveaparams.update(kwargs)
        return rveaparams

    def _run_interruption(self, population: "Population"):
        self.params["reference_vectors"].slow_interactive_adapt(
            self.params["ref_point"]
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['desdeo_emo',
 'desdeo_emo.EAs',
 'desdeo_emo.othertools',
 'desdeo_emo.population',
 'desdeo_emo.recombination',
 'desdeo_emo.selection',
 'desdeo_emo.surrogatemodels']

package_data = \
{'': ['*']}

install_requires = \
['desdeo-problem>=1.2.0,<2.0.0',
 'desdeo-tools>=1.4.1,<2.0.0',
 'plotly>=4.1,<5.0',
 'pyDOE>=0.3.8,<0.4.0']

setup_kwargs = {
    'name': 'desdeo-emo',
    'version': '1.2.1',
    'description': 'The python version reference vector guided evolutionary algorithm.',
    'long_description': '# desdeo-emo\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/industrial-optimization-group/desdeo-emo/master)\n\nThe evolutionary algorithms package within the `desdeo` framework.\n\nCode for the SoftwareX paper can be found in [this notebook](docs/notebooks/Using_EvoNN_for_optimization.ipynb).\n\nCurrently supported:\n* Multi-objective optimization with visualization and interaction support.\n* Preference is accepted as a reference point.\n* Surrogate modelling (neural networks and genetic trees) evolved via EAs.\n* Surrogate assisted optimization\n* Constraint handling using `RVEA`\n* IOPIS optimization using `RVEA` and `NSGA-III`\n\nCurrently _NOT_ supported:\n* Binary and integer variables.\n\nTo test the code, open the [binder link](https://mybinder.org/v2/gh/industrial-optimization-group/desdeo-emo/master) and read example.ipynb.\n\nRead the documentation [here](https://desdeo-emo.readthedocs.io/en/latest/)\n\n### Requirements\n* Python 3.7 (3.8 is NOT supported at the moment)\n* [Poetry dependency manager](https://github.com/sdispater/poetry): Only for developers\n\n### Installation process for normal users\n* Create a new virtual enviroment for the project\n* Run: `pip install desdeo_emo`\n\n### Installation process for developers\n* Download and extract the code or `git clone`\n* Create a new virtual environment for the project\n* Run `poetry install` inside the virtual environment shell.',
    'author': 'Bhupinder Saini',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<3.10',
}


setup(**setup_kwargs)



from collections.abc import Sequence

from typing import TYPE_CHECKING
import numpy as np
import pandas as pd
from pygmo import fast_non_dominated_sorting as nds
from pygmo import hypervolume as hv
from pygmo import non_dominated_front_2d as nd2

from tqdm import tqdm, tqdm_notebook

from pyrvea.Population.create_individuals import create_new_individuals

import plotly
import plotly.graph_objs as go

from pyrvea.OtherTools.plotlyanimate import animate_init_, animate_next_
from pyrvea.OtherTools.IsNotebook import IsNotebook
from pyrvea.Recombination import (
    biogp_xover,
    biogp_mutation,
    evodn2_xover_mutation,
    evonn_xover_mutation,
    bounded_polynomial_mutation,
    simulated_binary_crossover,
)


from desdeo_problem.Problem import MOProblem


class Population:
    """Define the population."""

    def __init__(
        self,
        problem: MOProblem,
        assign_type: str = "RandomDesign",
        pop_size=None,
        recombination_type=None,
        crossover_type="simulated_binary_crossover",
        mutation_type="bounded_polynomial_mutation",
        *args
    ):
        """Initialize the population.

        Parameters
        ----------
        problem : BaseProblem
            An object of the class Problem
        assign_type : str, optional
            Define the method of creation of population.
            If 'assign_type' is 'RandomDesign' the population is generated
            randomly. If 'assign_type' is 'LHSDesign', the population is
            generated via Latin Hypercube Sampling. If 'assign_type' is
            'custom', the population is imported from file. If assign_type
            is 'empty', create blank population.
            'EvoNN' and 'EvoDN2' will create neural networks or deep neural networks,
            respectively,
             for population .
        plotting : bool, optional
            (the default is True, which creates the plots)
        pop_size : int
            Population size
        recombination_type, crossover_type, mutation_type : str
            Recombination functions. If recombination_type is specified, crossover and
            mutation
            will be handled by the same function. If None, they are done separately.

        """
        self.assign_type = assign_type
        self.num_var = problem.n_of_variables
        self.lower_limits = np.asarray(problem.get_variable_lower_bounds())
        self.upper_limits = np.asarray(problem.get_variable_upper_bounds())
        self.hyp = 0
        self.non_dom = 0
        self.pop_size = pop_size
        # Fix to remove the following assumptions
        self.recombination_funcs = {
            "biogp_xover": biogp_xover,
            "biogp_mut": biogp_mutation,
            "evodn2_xover_mutation": evodn2_xover_mutation,
            "evonn_xover_mutation": evonn_xover_mutation,
            "bounded_polynomial_mutation": bounded_polynomial_mutation,
            "simulated_binary_crossover": simulated_binary_crossover,
        }
        self.crossover_type = crossover_type
        self.mutation_type = mutation_type
        self.recombination = self.recombination_funcs.get(recombination_type, None)
        if recombination_type is None:
            self.crossover = self.recombination_funcs.get(crossover_type, None)
            self.mutation = self.recombination_funcs.get(mutation_type, None)
        self.problem = problem
        self.filename = (
            problem.name + "_" + str(problem.n_of_objectives)
        )  # Used for plotting
        self.plotting = plotting
        self.individuals = []
        self.objectives = np.empty((0, self.problem.n_of_objectives), float)
        if problem.minimize is not None:
            self.fitness = self.objectives[:, self.problem.minimize]
            self.ideal_fitness = np.full((1, self.fitness.shape[1]), np.inf)
            self.worst_fitness = -1 * self.ideal_fitness
        else:
            self.fitness = np.empty((0, self.problem.num_of_objectives), float)
            self.ideal_fitness = np.full((1, self.problem.num_of_objectives), np.inf)
        self.worst_fitness = -1 * self.ideal_fitness
        self.constraint_violation = np.empty(
            (0, self.problem.num_of_constraints), float
        )
        self.archive = pd.DataFrame(
            columns=["generation", "decision_variables", "objective_values"]
        )

        if not assign_type == "empty":
            individuals = create_new_individuals(
                assign_type, problem, pop_size=self.pop_size
            )
            self.add(individuals)

        if self.plotting:
            self.figure = []
            self.plot_init_()

    def add(self, new_pop: list):
        """Evaluate and add individuals to the population. Update ideal and nadir point.

        Parameters
        ----------
        new_pop: list
            Decision variable values for new population.
        """
        for i in range(len(new_pop)):
            self.append_individual(new_pop[i])

        self.update_ideal_and_nadir()

    def append_individual(self, ind: np.ndarray):
        """Evaluate and add individual to the population.

        Parameters
        ----------
        ind: np.ndarray
        """

        self.individuals.append(ind)
        obj, CV, fitness = self.evaluate_individual(ind)
        self.objectives = np.vstack((self.objectives, obj))
        self.constraint_violation = np.vstack((self.constraint_violation, CV))
        self.fitness = np.vstack((self.fitness, fitness))

    def evaluate_individual(self, ind: np.ndarray):
        """Evaluate individual.

        Returns objective values, constraint violation, and fitness.

        Parameters
        ----------
        ind: np.ndarray
        """
        obj = self.problem.objectives(ind)
        CV = np.empty((0, self.problem.num_of_constraints), float)
        fitness = self.eval_fitness(obj)

        if self.problem.num_of_constraints:
            CV = self.problem.constraints(ind, obj)
            fitness = self.eval_fitness(obj)

        return obj, CV, fitness

    def eval_fitness(self, obj):
        """
        Calculate fitness based on objective values. Fitness = obj if minimized.
        """

        # fitness = self.objectives * self.problem.objs
        if self.problem.minimize is None:
            self.problem.minimize = [True] * self.problem.num_of_objectives
        else:
            assert len(self.problem.minimize) == self.problem.num_of_objectives

        fitness = np.asarray(obj)[np.asarray(self.problem.minimize)]

        return fitness

    def update_fitness(self):
        """Include or exclude objectives from fitness calculation.
        Problem.minimize should be a list of booleans of same length as the number of
        objectives.
        """
        self.fitness = self.objectives[:, self.problem.minimize]
        self.ideal_fitness = np.full((1, self.fitness.shape[1]), np.inf)
        self.worst_fitness = -1 * self.ideal_fitness
        self.update_ideal_and_nadir()

    def delete(self, indices, preserve=False):
        """Remove from population individuals which are in indices if preserve=False,
        otherwise preserve them and remove all others.

        Parameters
        ----------
        indices: array_like
            Indices of individuals to keep or delete.
        preserve: bool
            Whether to delete individuals at indices from current population, or
            preserve them and delete others.
        """

        mask = np.ones(len(self.individuals), dtype=bool)
        mask[indices] = False

        new_pop = np.array(self.individuals)[mask]
        deleted_pop = np.array(self.individuals)[~mask]

        new_obj = self.objectives[mask]
        deleted_obj = self.objectives[~mask]

        new_fitness = self.fitness[mask]
        deleted_fitness = self.fitness[~mask]

        if len(self.constraint_violation) > 0:
            new_cv = self.constraint_violation[mask]
            deleted_cv = self.constraint_violation[~mask]
        else:
            deleted_cv = self.constraint_violation
            new_cv = self.constraint_violation

        if not preserve:
            self.individuals = list(new_pop)
            self.objectives = new_obj
            self.fitness = new_fitness
            self.constraint_violation = new_cv

        else:
            self.individuals = list(deleted_pop)
            self.objectives = deleted_obj
            self.fitness = deleted_fitness
            self.constraint_violation = deleted_cv

    def evolve(self, EA: "BaseEA" = None, ea_parameters: dict = None):
        """Evolve the population with interruptions.

        Evolves the population based on the EA sent by the user.

        Parameters
        ----------
        EA: "BaseEA"
            Should be a derivative of BaseEA (Default value = None)
        ea_parameters: dict
            Contains the parameters needed by EA (Default value = None)

        """
        ##################################
        # To determine whether running in console or in notebook. Used for TQDM.
        # TQDM will be removed in future generations as number of iterations can vary

        if IsNotebook():
            progressbar = tqdm_notebook
        else:
            progressbar = tqdm
        ####################################
        # A basic evolution cycle. Will be updated to optimize() in future versions.
        ea = EA(self, ea_parameters)
        iterations = ea.params["iterations"]

        if self.plotting:
            self.plot_objectives()  # Figure was created in init
        for i in progressbar(range(iterations), desc="Iteration"):
            ea._run_interruption(self)
            ea._next_iteration(self)
            if self.plotting:
                self.plot_objectives()

    def mate(self, mating_pop=None, params=None):
        """Conduct crossover and mutation over the population.

        """

        if self.recombination is not None:
            offspring = self.recombination.mate(
                mating_pop,
                self.individuals,
                params,
                crossover_type=self.crossover_type,
                mutation_type=self.mutation_type,
            )
        else:
            offspring = self.crossover.mate(mating_pop, self.individuals, params)
            self.mutation.mutate(
                offspring,
                self.individuals,
                params,
                self.lower_limits,
                self.upper_limits,
            )

        return offspring

    def plot_init_(self):
        """Initialize animation object. Return figure"""
        obj = self.objectives
        self.figure = animate_init_(obj, self.filename + ".html")
        return self.figure

    def plot_objectives(self, iteration: int = None):
        """Plot the objective values of individuals.

        Parameters
        ----------
        iteration: int
            Iteration count.
        """
        obj = self.objectives
        self.figure = animate_next_(
            obj, self.figure, self.filename + ".html", iteration
        )

    def plot_pareto(self, name, show_all=False):
        """Plot the pareto front. REMOVE THIS IN THE FUTURE.

        Parameters
        ----------
        name : str
            Name to append to the plot filename.
        show_all : bool
            Show all solutions, including those not on the pareto front.

        """
        if name is None:
            name = self.problem.name

        ndf = self.non_dominated()
        # pareto = self.objectives[ndf][self.objectives[ndf].min(axis=1) >= 0, :]
        pareto = self.objectives[ndf]
        pareto_pop = np.asarray(self.individuals)[ndf].tolist()

        for idx, x in enumerate(pareto_pop):

            for i, y in enumerate(x):
                x[i] = "x" + str(i + 1) + ": " + str(y) + "<br>"
            x.insert(0, "Model " + str(idx))

        trace0 = go.Scatter(
            x=pareto[:, 0],
            y=pareto[:, 1],
            text=pareto_pop,
            hoverinfo="text",
            mode="markers+lines",
        )

        if show_all:
            trace1 = go.Scatter(
                x=self.objectives[:, 0], y=self.objectives[:, 1], mode="markers"
            )

            data = [trace0, trace1]
        else:
            data = [trace0]

        layout = go.Layout(xaxis=dict(title="f1"), yaxis=dict(title="f2"))
        plotly.offline.plot(
            {"data": data, "layout": layout},
            filename=name + "pareto" + ".html",
            auto_open=True,
        )

    def hypervolume(self, ref_point):
        """Calculate hypervolume. Uses package pygmo. Add checks to prevent errors.

        Parameters
        ----------
        ref_point

        Returns
        -------

        """
        non_dom = self.non_dom
        if not isinstance(ref_point, (Sequence, np.ndarray)):
            num_obj = non_dom.shape[1]
            ref_point = [ref_point] * num_obj
        non_dom = non_dom[np.all(non_dom < ref_point, axis=1), :]
        hyp = hv(non_dom)
        self.hyp = hyp.compute(ref_point)
        return self.hyp

    def non_dominated(self):
        """Fix this. check if nd2 and nds mean the same thing"""
        obj = self.objectives
        num_obj = obj.shape[1]
        if num_obj == 2:
            non_dom_front = nd2(obj)
        else:
            non_dom_front = nds(obj)
        if isinstance(non_dom_front, tuple):
            self.non_dom = self.objectives[non_dom_front[0][0]]
        elif isinstance(non_dom_front, np.ndarray):
            self.non_dom = self.objectives[non_dom_front]
        else:
            print("Non Dom error Line 285 in population.py")
        return non_dom_front

    def update_ideal_and_nadir(self, new_objective_vals: list = None):
        """Updates self.ideal and self.nadir in the fitness space.

        Uses the entire population if new_objective_vals is none.

        Parameters
        ----------
        new_objective_vals : list, optional
            Objective values for a newly added individual (the default is None, which
            calculated the ideal and nadir for the entire population.)

        """
        if new_objective_vals is None:
            check_ideal_with = self.fitness
        else:
            check_ideal_with = new_objective_vals
        self.ideal_fitness = np.amin(
            np.vstack((self.ideal_fitness, check_ideal_with)), axis=0
        )
        self.worst_fitness = np.amax(
            np.vstack((self.worst_fitness, check_ideal_with)), axis=0
        )



from typing import Dict

import numpy as np
from scipy.spatial.distance import cdist as s_cdist

from numpy.random import permutation
from scipy.spatial import distance_matrix
from desdeo_emo.EAs.BaseEA import BaseDecompositionEA
from desdeo_emo.population.Population import Population
from desdeo_problem.Problem import MOProblem

from desdeo_emo.selection import tournament_select
from desdeo_emo.selection.MOEAD_select import MOEAD_select
from desdeo_emo.recombination.BoundedPolynomialMutation import BP_mutation
from desdeo_emo.recombination.SimulatedBinaryCrossover import SBX_xover

from desdeo_tools.scalarization import MOEADSF
from desdeo_tools.scalarization.MOEADSF import Tchebycheff


class MOEA_D(BaseDecompositionEA):
    """Python implementation of MOEA/D

    .. Q. Zhang and H. Li, "MOEA/D: A Multiobjective Evolutionary Algorithm Based on Decomposition," 
    in IEEE Transactions on Evolutionary Computation, vol. 11, no. 6, pp. 712-731, Dec. 2007, doi: 10.1109/TEVC.2007.892759.

    Parameters
    ----------
    problem: MOProblem
    	The problem class object specifying the details of the problem.
    scalarization_function: MOEADSF
    	The scalarization function to compare the solutions. Some implementations 
        can be found in desdeo-tools/scalarization/MOEADSF. By default it uses the
        Tchebycheff function.
    n_neighbors: int, optional
    	Number of reference vectors considered in the neighborhoods creation. The default 
    	number is 20.
    population_params: Dict, optional
    	The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population: Population, optional
    	An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    lattice_resolution: int, optional
    	The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    n_parents: int, optional
    	Number of individuals considered for the generation of offspring solutions. The default
    	option is 2.
    a_priori: bool, optional
    	A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact: bool, optional
    	A bool variable defining whether interactive preference is to be used or
        not. By default False
    use_surrogates: bool, optional
    	A bool variable defining whether surrogate problems are to be used or
        not. By default False
    n_iterations: int, optional
     	The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter: int, optional
    	The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations: int, optional
    	Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    """

    def __init__(  # parameters of the class
        self,
        problem: MOProblem,
        scalarization_function: MOEADSF = Tchebycheff(),
        n_neighbors: int = 20,
        population_params: Dict = None,
        initial_population: Population = None,
        lattice_resolution: int = None,
        use_repair: bool = True,
        n_parents: int = 2,
        a_priori: bool = False,
        interact: bool = False,
        use_surrogates: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
    ):
        super().__init__(  # parameters for decomposition based approach
            problem=problem,
            population_size=None,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            a_priori=a_priori,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
        )
        self.population_size = self.population.pop_size
        self.problem = problem
        self.scalarization_function = scalarization_function
        self.n_neighbors = n_neighbors

        self.use_repair = use_repair
        self.n_parents = n_parents
        self.population.mutation = BP_mutation(
            problem.get_variable_lower_bounds(),
            problem.get_variable_upper_bounds(),
            0.5,
            20,
        )
        self.population.recombination = SBX_xover(1.0, 20)

        selection_operator = MOEAD_select(
            self.population, SF_type=self.scalarization_function
        )
        self.selection_operator = selection_operator
        # Compute the distance between each pair of reference vectors
        distance_matrix_vectors = distance_matrix(
            self.reference_vectors.values, self.reference_vectors.values
        )
        # Get the closest vectors to obtain the neighborhoods
        self.neighborhoods = np.argsort(
            distance_matrix_vectors, axis=1, kind="quicksort"
        )[:, :n_neighbors]
        self.population.update_ideal()
        self._ideal_point = self.population.ideal_objective_vector

    def _next_gen(self):
        # For each individual from the population
        for i in range(self.population_size):
            # Consider only the individuals of the current neighborhood
            # for parent selection
            current_neighborhood = self.neighborhoods[i, :]
            selected_parents = current_neighborhood[permutation(self.n_neighbors)][
                : self.n_parents
            ]

            offspring = self.population.recombination.do(
                self.population.individuals, selected_parents
            )
            offspring = self.population.mutation.do(offspring)
            # Apply genetic operators over two random individuals
            # offspring = self.population.mate(selected_parents)
            offspring = np.array(offspring[0, :])

            # Repair the solution if it is needed
            if self.use_repair:
                offspring = self.population.repair(offspring)

            # Evaluate the offspring using the objective function
            results_off = self.problem.evaluate(offspring, self.use_surrogates)
            offspring_fx = results_off.objectives
            self._function_evaluation_count += 1

            # Update the ideal point
            self._ideal_point = np.min(
                np.vstack([self._ideal_point, offspring_fx]), axis=0
            )

            # Replace individuals with a worse SF value than the offspring
            selected = self._select(current_neighborhood, offspring_fx)

            self.population.replace(selected, offspring, results_off)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1

    def _select(self, current_neighborhood, offspring_fx) -> list:
        return self.selection_operator.do(
            self.population,
            self.reference_vectors,
            self._ideal_point,
            current_neighborhood,
            offspring_fx,
        )




from typing import Dict, Union, List

from desdeo_emo.EAs.BaseEA import eaError
from desdeo_emo.EAs.BaseEA import BaseDecompositionEA, BaseEA
from desdeo_emo.EAs.RVEA import RVEA
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.IOPIS_APD import IOPIS_APD_Select
from desdeo_emo.selection.IOPIS_NSGAIII import IOPIS_NSGAIII_select
from desdeo_problem.Problem import MOProblem
from desdeo_tools.scalarization import StomASF, PointMethodASF, AugmentedGuessASF
from desdeo_emo.othertools.ReferenceVectors import ReferenceVectors
from desdeo_tools.interaction import (
    ReferencePointPreference,
    validate_ref_point_with_ideal_and_nadir,
)
import numpy as np
import pandas as pd


class BaseIOPISDecompositionEA(BaseDecompositionEA, BaseEA):
    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        lattice_resolution: int = None,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        use_surrogates: bool = False,
    ):
        a_priori: bool = True
        interact: bool = True
        if problem.ideal is None or problem.nadir is None:
            msg = (
                f"The problem instance should contain the information about ideal and "
                f"nadir point."
            )
            raise eaError(msg)

        BaseEA.__init__(
            self=self,
            a_priori=a_priori,
            interact=interact,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            use_surrogates=use_surrogates,
        )

        scalarization_methods = [
            StomASF(ideal=problem.ideal * problem._max_multiplier),
            # PointMethodASF(
            #     nadir=problem.nadir * problem._max_multiplier,
            #     ideal=problem.ideal * problem._max_multiplier,
            # ),
            AugmentedGuessASF(
                nadir=problem.nadir * problem._max_multiplier,
                ideal=problem.ideal * problem._max_multiplier,
                indx_to_exclude=[],
            ),
        ]
        if lattice_resolution is None:
            lattice_res_options = [49, 13, 7, 5, 4, 3, 3, 3, 3]
            if len(scalarization_methods) < 11:
                lattice_resolution = lattice_res_options[len(scalarization_methods) - 2]
            else:
                lattice_resolution = 3
        reference_vectors = ReferenceVectors(
            lattice_resolution=lattice_resolution,
            number_of_objectives=len(scalarization_methods),
        )
        population_size = reference_vectors.number_of_vectors
        population = Population(problem, population_size, population_params)

        self.reference_vectors = reference_vectors
        self.scalarization_methods = scalarization_methods

        if initial_population is not None:
            #  Population should be compatible.
            self.population = initial_population  # TODO put checks here.
        elif initial_population is None:
            if population_size is None:
                population_size = self.reference_vectors.number_of_vectors
            self.population = Population(
                problem, population_size, population_params, use_surrogates
            )
            self._function_evaluation_count += population_size
        self._ref_vectors_are_focused: bool = False

    def manage_preferences(self, preference=None):
        """Run the interruption phase of EA.

        Use this phase to make changes to RVEA.params or other objects.
        Updates Reference Vectors (adaptation), conducts interaction with the user.
        """
        if preference is None:
            msg = "Giving preferences is mandatory"
            raise eaError(msg)

        if not isinstance(preference, ReferencePointPreference):
            msg = (
                f"Wrong object sent as preference. Expected type = "
                f"{type(ReferencePointPreference)} or None\n"
                f"Recieved type = {type(preference)}"
            )
            raise eaError(msg)

        if preference.request_id != self._interaction_request_id:
            msg = (
                f"Wrong preference object sent. Expected id = "
                f"{self._interaction_request_id}.\n"
                f"Recieved id = {preference.request_id}"
            )
            raise eaError(msg)

        refpoint = preference.response.values * self.population.problem._max_multiplier
        self._preference = refpoint
        scalarized_space_fitness = np.asarray(
            [
                scalar(self.population.fitness, self._preference)
                for scalar in self.scalarization_methods
            ]
        ).T
        self.reference_vectors.adapt(scalarized_space_fitness)
        self.reference_vectors.neighbouring_angles()

    def request_preferences(self) -> Union[None, ReferencePointPreference]:
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=self.population.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self.population.problem._max_multiplier
        dimensions_data.loc["ideal"] = self.population.ideal_objective_vector
        dimensions_data.loc["nadir"] = self.population.nadir_objective_vector
        message = (
            f"Provide a reference point worse than to the ideal point and better than"
            f" the nadir point.\n"
            f"Ideal point: \n{dimensions_data.loc['ideal']}\n"
            f"Nadir point: \n{dimensions_data.loc['nadir']}\n"
            f"The reference point will be used to create scalarization functions in "
            f"the preferred region.\n"
        )
        interaction_priority = "required"
        self._interaction_request_id = np.random.randint(0, 1e7)
        return ReferencePointPreference(
            dimensions_data=dimensions_data,
            message=message,
            interaction_priority=interaction_priority,
            preference_validator=validate_ref_point_with_ideal_and_nadir,
            request_id=self._interaction_request_id,
        )

    def _select(self) -> List:
        return self.selection_operator.do(
            self.population, self.reference_vectors, self._preference
        )


class IOPIS_RVEA(BaseIOPISDecompositionEA, RVEA):
    """The python version reference vector guided evolutionary algorithm.

    Most of the relevant code is contained in the super class. This class just assigns
    the APD selection operator to BaseDecompositionEA.

    NOTE: The APD function had to be slightly modified to accomodate for the fact that
    this version of the algorithm is interactive, and does not have a set termination
    criteria. There is a time component in the APD penalty function formula of the type:
    (t/t_max)^alpha. As there is no set t_max, the formula has been changed. See below,
    the documentation for the argument: penalty_time_component

    See the details of RVEA in the following paper

    R. Cheng, Y. Jin, M. Olhofer and B. Sendhoff, A Reference Vector Guided
    Evolutionary Algorithm for Many-objective Optimization, IEEE Transactions on
    Evolutionary Computation, 2016

    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    alpha : float, optional
        The alpha parameter in the APD selection mechanism. Read paper for details.
    lattice_resolution : int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    penalty_time_component: Union[str, float], optional
        The APD formula had to be slightly changed.
        If penalty_time_component is a float between [0, 1], (t/t_max) is replaced by
        that constant for the entire algorithm.
        If penalty_time_component is "original", the original intent of the paper is
        followed and (t/t_max) is calculated as
        (current generation count/total number of generations).
        If penalty_time_component is "function_count", (t/t_max) is calculated as
        (current function evaluation count/total number of function evaluations)
        If penalty_time_component is "interactive", (t/t_max)  is calculated as
        (Current gen count within an iteration/Total gen count within an iteration).
        Hence, time penalty is always zero at the beginning of each iteration, and one
        at the end of each iteration.
        Note: If the penalty_time_component ever exceeds one, the value one is used as
        the penalty_time_component.
        If no value is provided, an appropriate default is selected.
        If `interact` is true, penalty_time_component is "interactive" by default.
        If `interact` is false, but `total_function_evaluations` is provided,
        penalty_time_component is "function_count" by default.
        If `interact` is false, but `total_function_evaluations` is not provided,
        penalty_time_component is "original" by default.
    """

    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        alpha: float = None,
        lattice_resolution: int = None,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        time_penalty_component: Union[str, float] = None,
        use_surrogates: bool = False,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            use_surrogates=use_surrogates,
        )
        self.time_penalty_component = time_penalty_component
        time_penalty_component_options = ["original", "function_count", "interactive"]
        if time_penalty_component is None:
            if self.interact is True:
                time_penalty_component = "interactive"
            elif total_function_evaluations > 0:
                time_penalty_component = "function_count"
            else:
                time_penalty_component = "original"
        if not (type(time_penalty_component) is float or str):
            msg = (
                f"type(time_penalty_component) should be float or str"
                f"Provided type: {type(time_penalty_component)}"
            )
            eaError(msg)
        if type(time_penalty_component) is float:
            if (time_penalty_component <= 0) or (time_penalty_component >= 1):
                msg = (
                    f"time_penalty_component should either be a float in the range"
                    f"[0, 1], or one of {time_penalty_component_options}.\n"
                    f"Provided value = {time_penalty_component}"
                )
                eaError(msg)
            time_penalty_function = self._time_penalty_constant
        if type(time_penalty_component) is str:
            if time_penalty_component == "original":
                time_penalty_function = self._time_penalty_original
            elif time_penalty_component == "function_count":
                time_penalty_function = self._time_penalty_function_count
            elif time_penalty_component == "interactive":
                time_penalty_function = self._time_penalty_interactive
            else:
                msg = (
                    f"time_penalty_component should either be a float in the range"
                    f"[0, 1], or one of {time_penalty_component_options}.\n"
                    f"Provided value = {time_penalty_component}"
                )
                eaError(msg)
        self.time_penalty_function = time_penalty_function
        self.alpha = alpha
        selection_operator = IOPIS_APD_Select(
            self.time_penalty_function, self.scalarization_methods, self.alpha
        )
        self.selection_operator = selection_operator

    def _time_penalty_constant(self):
        """Returns the constant time penalty value.
        """
        return self.time_penalty_component

    def _time_penalty_original(self):
        """Calculates the appropriate time penalty value, by the original formula.
        """
        return self._current_gen_count / self.total_gen_count

    def _time_penalty_interactive(self):
        """Calculates the appropriate time penalty value.
        """
        return self._gen_count_in_curr_iteration / self.n_gen_per_iter

    def _time_penalty_function_count(self):
        """Calculates the appropriate time penalty value.
        """
        return self._function_evaluation_count / self.total_function_evaluations


class IOPIS_NSGAIII(BaseIOPISDecompositionEA):
    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        lattice_resolution: int = None,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        use_surrogates: bool = False,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            use_surrogates=use_surrogates,
        )
        self.selection_operator = IOPIS_NSGAIII_select(
            self.scalarization_methods, self.population
        )




from typing import Dict

from desdeo_emo.EAs.BaseEA import BaseDecompositionEA
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.NSGAIII_select import NSGAIII_select
from desdeo_problem.Problem import MOProblem


class NSGAIII(BaseDecompositionEA):
    """Python Implementation of NSGA-III. Based on the pymoo package.

    Most of the relevant code is contained in the super class. This class just assigns
    the NSGAIII selection operator to BaseDecompositionEA.

    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    lattice_resolution : int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    selection_type : str, optional
        One of ["mean", "optimistic", "robust"]. To be used in data-driven optimization.
        To be used only with surrogate models which return an "uncertainity" factor.
        Using "mean" is equivalent to using the mean predicted values from the surrogate
        models and is the default case.
        Using "optimistic" results in using (mean - uncertainity) values from the
        the surrogate models as the predicted value (in case of minimization). It is
        (mean + uncertainity for maximization).
        Using "robust" is the opposite of using "optimistic".
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    """

    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        n_survive: int = None,
        initial_population: Population = None,
        lattice_resolution: int = None,
        selection_type: str = None,
        a_priori: bool = False,
        interact: bool = False,
        use_surrogates: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            a_priori=a_priori,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
        )
        self.selection_type = selection_type
        selection_operator = NSGAIII_select(
            self.population, n_survive, selection_type=selection_type
        )
        self.selection_operator = selection_operator



from typing import Dict, Type, Union, Tuple

import numpy as np
import pandas as pd

from desdeo_emo.othertools.ReferenceVectors import ReferenceVectors
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.SelectionBase import SelectionBase
from desdeo_problem.Problem import MOProblem
from desdeo_tools.interaction import (
    SimplePlotRequest,
    ReferencePointPreference,
    PreferredSolutionPreference,
    NonPreferredSolutionPreference,
    BoundPreference,
    validate_ref_point_data_type,
    validate_ref_point_dimensions,
    validate_ref_point_with_ideal,
)


class eaError(Exception):
    """Raised when an error related to EA occurs"""


class BaseEA:
    """This class provides the basic structure for Evolutionary algorithms."""

    def __init__(
        self,
        a_priori: bool = False,
        interact: bool = False,
        selection_operator: Type[SelectionBase] = None,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        use_surrogates: bool = False,
    ):
        """Initialize EA here. Set up parameters, create EA specific objects."""
        self.a_priori: bool = a_priori
        self.interact: bool = interact
        self.n_iterations: int = n_iterations
        self.n_gen_per_iter: int = n_gen_per_iter
        self.total_gen_count: int = n_gen_per_iter * n_iterations
        self.total_function_evaluations = total_function_evaluations
        self.selection_operator = selection_operator
        self.use_surrogates: bool = use_surrogates
        # Internal counters and state trackers
        self._iteration_counter: int = 0
        self._gen_count_in_curr_iteration: int = 0
        self._current_gen_count: int = 0
        self._function_evaluation_count: int = 0

    def _next_gen(self):
        """Run one generation of an EA. Change nothing about the parameters."""

    def iterate(self, preference=None) -> Tuple:
        """Run one iteration of EA.

        One iteration consists of a constant or variable number of
        generations. This method leaves EA.params unchanged, except the current
        iteration count and gen count.
        """
        self.manage_preferences(preference)
        self._gen_count_in_curr_iteration = 0
        while self.continue_iteration():
            self._next_gen()
        self._iteration_counter += 1
        return self.requests()

    def continue_iteration(self):
        """Checks whether the current iteration should be continued or not."""
        return (
            self._gen_count_in_curr_iteration < self.n_gen_per_iter
            and self.check_FE_count()
        )

    def continue_evolution(self) -> bool:
        """Checks whether the current iteration should be continued or not."""
        return self._iteration_counter < self.n_iterations and self.check_FE_count()

    def check_FE_count(self) -> bool:
        """Checks whether termination criteria via function evaluation count has been
            met or not.

        Returns:
            bool: True is function evaluation count limit NOT met.
        """
        if self.total_function_evaluations == 0:
            return True
        elif self._function_evaluation_count <= self.total_function_evaluations:
            return True
        return False

    def manage_preferences(self, preference=None):
        """Run the interruption phase of EA.

        Use this phase to make changes to RVEA.params or other objects.
        Updates Reference Vectors (adaptation), conducts interaction with the user.
        """
        pass

    def requests(self) -> Tuple:
        pass


class BaseDecompositionEA(BaseEA):
    """The Base class for decomposition based EAs.

    This class contains most of the code to set up the parameters and operators.
    It also contains the logic of a simple decomposition EA.

    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    selection_operator : Type[SelectionBase], optional
        The selection operator to be used by the EA, by default None.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    lattice_resolution : int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    """

    def __init__(
        self,
        problem: MOProblem,
        selection_operator: Type[SelectionBase] = None,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        a_priori: bool = False,
        interact: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        lattice_resolution: int = None,
        use_surrogates: bool = False,
    ):
        super().__init__(
            a_priori=a_priori,
            interact=interact,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            selection_operator=selection_operator,
            use_surrogates=use_surrogates,
        )
        if lattice_resolution is None:
            lattice_res_options = [49, 13, 7, 5, 4, 3, 3, 3, 3]
            if problem.n_of_objectives < 11:
                lattice_resolution = lattice_res_options[problem.n_of_objectives - 2]
            else:
                lattice_resolution = 3
        self.reference_vectors = ReferenceVectors(
            lattice_resolution, problem.n_of_objectives
        )
        if initial_population is not None:
            #  Population should be compatible.
            self.population = initial_population  # TODO put checks here.
        elif initial_population is None:
            if population_size is None:
                population_size = self.reference_vectors.number_of_vectors
            self.population = Population(
                problem, population_size, population_params, use_surrogates
            )
            self._function_evaluation_count += population_size
        self._ref_vectors_are_focused: bool = False
        # print("Using BaseDecompositionEA init")

    def _next_gen(self):
        """Run one generation of decomposition based EA. Intended to be used by
        next_iteration.
        """
        offspring = self.population.mate()  # (params=self.params)
        self.population.add(offspring, self.use_surrogates)
        selected = self._select()
        self.population.keep(selected)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1
        self._function_evaluation_count += offspring.shape[0]

    def manage_preferences(self, preference=None):
        """Run the interruption phase of EA.

        Use this phase to make changes to RVEA.params or other objects.
        Updates Reference Vectors (adaptation), conducts interaction with the user.
        """
        if not isinstance(
            preference,
            (
                ReferencePointPreference,
                PreferredSolutionPreference,
                NonPreferredSolutionPreference,
                BoundPreference,
                type(None),
            ),
        ):
            msg = (
                f"Wrong object sent as preference. Expected type = "
                f"{type(ReferencePointPreference)}\n"
                f"{type(PreferredSolutionPreference)}\n"
                f"{type(NonPreferredSolutionPreference)}\n"
                f"{type(BoundPreference)} or None\n"
                f"Recieved type = {type(preference)}"
            )
            raise eaError(msg)
        if preference is not None:
            if preference.request_id != self._interaction_request_id:
                msg = (
                    f"Wrong preference object sent. Expected id = "
                    f"{self._interaction_request_id}.\n"
                    f"Recieved id = {preference.request_id}"
                )
                raise eaError(msg)
        if preference is None and not self._ref_vectors_are_focused:
            self.reference_vectors.adapt(self.population.fitness)
        if isinstance(preference, ReferencePointPreference):
            ideal = self.population.ideal_fitness_val
            refpoint = (
                preference.response.values * self.population.problem._max_multiplier
            )
            refpoint = refpoint - ideal
            norm = np.sqrt(np.sum(np.square(refpoint)))
            refpoint = refpoint / norm
            self.reference_vectors.iteractive_adapt_3(refpoint)
            self.reference_vectors.add_edge_vectors()
        elif isinstance(preference, PreferredSolutionPreference):
            self.reference_vectors.interactive_adapt_1(
                z=self.population.objectives[preference.response],
                n_solutions=np.shape(self.population.objectives)[0],
            )
            self.reference_vectors.add_edge_vectors()
        elif isinstance(preference, NonPreferredSolutionPreference):
            self.reference_vectors.interactive_adapt_2(
                z=self.population.objectives[preference.response],
                n_solutions=np.shape(self.population.objectives)[0],
            )
            self.reference_vectors.add_edge_vectors()
        elif isinstance(preference, BoundPreference):
            self.reference_vectors.interactive_adapt_4(preference.response)
            self.reference_vectors.add_edge_vectors()
        self.reference_vectors.neighbouring_angles()

    def _select(self) -> list:
        """Describe a selection mechanism. Return indices of selected
        individuals.

        Returns
        -------
        list
            List of indices of individuals to be selected.
        """
        return self.selection_operator.do(self.population, self.reference_vectors)

    def request_plot(self) -> SimplePlotRequest:
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=self.population.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self.population.problem._max_multiplier
        dimensions_data.loc["ideal"] = self.population.ideal_objective_vector
        dimensions_data.loc["nadir"] = self.population.nadir_objective_vector
        data = pd.DataFrame(
            self.population.objectives, columns=self.population.problem.objective_names
        )
        return SimplePlotRequest(
            data=data, dimensions_data=dimensions_data, message="Objective Values"
        )

    def request_preferences(
        self
    ) -> Union[
        None,
        Tuple[
            PreferredSolutionPreference,
            NonPreferredSolutionPreference,
            ReferencePointPreference,
            BoundPreference,
        ],
    ]:

        if self.a_priori is False and self.interact is False:
            return
        if (
            self.a_priori is True
            and self.interact is False
            and self._iteration_counter > 0
        ):
            return
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=self.population.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self.population.problem._max_multiplier
        dimensions_data.loc["ideal"] = self.population.ideal_objective_vector
        dimensions_data.loc["nadir"] = self.population.nadir_objective_vector
        message = (
            "Please provide preferences. There is four ways to do this. You can either:\n\n"
            "\t1: Select preferred solution(s)\n"
            "\t2: Select non-preferred solution(s)\n"
            "\t3: Specify a reference point worse than or equal to the ideal point\n"
            "\t4: Specify desired ranges for objectives.\n\n"
            "In case you choose \n\n"
            "1, please specify index/indices of preferred solutions in a numpy array (indexing starts from 0).\n"
            "For example: \n"
            "\tnumpy.array([1]), for choosing the solutions with index 1.\n"
            "\tnumpy.array([2, 4, 5, 16]), for choosing the solutions with indices 2, 4, 5, and 16.\n\n"
            "2, please specify index/indices of non-preferred solutions in a numpy array (indexing starts from 0).\n"
            "For example: \n"
            "\tnumpy.array([3]), for choosing the solutions with index 3.\n"
            "\tnumpy.array([1, 2]), for choosing the solutions with indices 1 and 2.\n\n"
            "3, please provide a reference point worse than or equal to the ideal point:\n\n"
            f"{dimensions_data.loc['ideal']}\n"
            f"The reference point will be used to focus the reference vectors towards "
            f"the preferred region.\n"
            f"If a reference point is not provided, the previous state of the reference"
            f" vectors is used.\n"
            f"If the reference point is the same as the ideal point, "
            f"the reference vectors are spread uniformly in the objective space.\n\n"
            "4, please specify desired lower and upper bound for each objective, starting from \n"
            "the first objective and ending with the last one. Please specify the bounds as a numpy array containing \n"
            "lists, so that the first item of list is the lower bound and the second the upper bound, for each \n"
            "objective. \n"
            "\tFor example: numpy.array([[1, 2], [2, 5], [0, 3.5]]), for problem with three "
            "objectives.\n"
            f"Ideal vector: \n{dimensions_data.loc['ideal']}\n"
            f"Nadir vector: \n{dimensions_data.loc['nadir']}."
        )

        def validator(dimensions_data: pd.DataFrame, reference_point: pd.DataFrame):
            validate_ref_point_dimensions(dimensions_data, reference_point)
            validate_ref_point_data_type(reference_point)
            validate_ref_point_with_ideal(dimensions_data, reference_point)
            return

        interaction_priority = "recommended"
        self._interaction_request_id = np.random.randint(0, 1e9)

        # return multiple preference-requests, user decides with request will (s)he respond to by using an index.
        return (
            PreferredSolutionPreference(
                n_solutions=self.population.objectives.shape[0],
                message=message,
                interaction_priority=interaction_priority,
                request_id=self._interaction_request_id,
            ),
            NonPreferredSolutionPreference(
                n_solutions=self.population.objectives.shape[0],
                message=None,
                interaction_priority=interaction_priority,
                request_id=self._interaction_request_id,
            ),
            ReferencePointPreference(
                dimensions_data=dimensions_data,
                message=None,
                interaction_priority=interaction_priority,
                preference_validator=validator,
                request_id=self._interaction_request_id,
            ),
            BoundPreference(
                dimensions_data=dimensions_data,
                n_objectives=self.population.problem.n_of_objectives,
                message=None,
                interaction_priority=interaction_priority,
                request_id=self._interaction_request_id,
            ),
        )

    def requests(self) -> Tuple:
        return (self.request_plot(), self.request_preferences())





def IsNotebook() -> bool:
    """Checks if the current environment is a Jupyter Notebook or a console.

    Returns
    -------
    bool
        True if notebook. False if console
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False
